#include <iostream>
#include <fstream>
#include <iterator>
#include <string>
#include <list>
#include <algorithm>
#include <functional>
#include <clocale>
#include <cctype>
#include <tchar.h>
#include <codecvt>

#ifdef NDEBUG
#include <windows.h>
#endif

using namespace std;

typedef basic_string<TCHAR> tstring;
typedef list<tstring> commands;

#ifdef UNICODE
#define terr wcerr

#define MAX_CODE 0x10ffff
wstring utf8_to_wstring(string const& src)
{
	wstring_convert<codecvt_utf8_utf16<wchar_t, MAX_CODE, consume_header>> converter;
	return converter.from_bytes(src);
}

string wstring_to_utf8(wstring const& src, bool addBom = false)
{
	if (addBom) {
		wstring_convert<codecvt_utf8_utf16<wchar_t, MAX_CODE, generate_header>> converter;
		return converter.to_bytes(src);
	}
	else {
		wstring_convert<codecvt_utf8_utf16<wchar_t>> converter;
		return converter.to_bytes(src);
	}
}
#else
#define terr cerr
#define tifstream ifstream
#define tofstream ofstream
#endif

void add_double_quote(tstring& st)
{
	if (*st.begin() != _T('"')) {
		st.insert(0, 1, _T('"'));
	}
	if (*st.rbegin() != _T('"')) {
		st.push_back(_T('"'));
	}
}

void remove_double_quote(tstring& st)
{
	remove_if(st.begin(), st.end(),
			  bind1st(equal_to<tstring::value_type>(), _T('"')));
}

bool load(const TCHAR* filename, const tstring& cmd, commands& cmdlist, tstring* replace)
{
	ifstream file(filename);
#ifdef UNICODE
	string buf8;
#endif
	tstring buf;
	bool find = false;
	
	if (cmd.empty()) {
#ifdef UNICODE
		while (getline(file, buf8)) {
			buf = utf8_to_wstring(buf8);
#else
		while (getline(file, buf)) {
#endif
			cmdlist.push_back(buf);
		}
		return find;
	}
#ifdef UNICODE
	while (getline(file, buf8)) {
		buf = utf8_to_wstring(buf8);
#else
	while (getline(file, buf)) {
#endif
		if (buf.length() < cmd.length() ||
			buf.compare(buf.length() - cmd.length(), cmd.length(), cmd) != 0 ||
			(buf.length() > cmd.length() && 
				(! isspace(buf[buf.length() - cmd.length() - 1]) &&
					buf[buf.length() - cmd.length() - 1] != _T('"')))) {
			cmdlist.push_back(buf);
		} else {
#ifndef NDEBUG
			terr << _T("'") << buf << _T("'") << endl;
#endif
			if (replace) {
				replace->swap(buf);
			}
			find = true;
		}
	}
	return find;
}

void save(const TCHAR* filename, const commands& cmdlist)
{
	ofstream file(filename);
#ifndef NDEBUG
	copy(cmdlist.begin(), cmdlist.end(), ostream_iterator<tstring, TCHAR>(terr, _T("\n")));
#endif
#ifdef UNICODE
	list<string> cmdlist8;
	list<tstring>::const_iterator it = cmdlist.begin();
	cmdlist8.push_back(wstring_to_utf8(*it++, true));
	while (it != cmdlist.end()) {
		cmdlist8.push_back(wstring_to_utf8(*it++));
	}
	copy(cmdlist8.begin(), cmdlist8.end(), ostream_iterator<string, char>(file, "\n"));
#else
	copy(cmdlist.begin(), cmdlist.end(), ostream_iterator<string, char>(file, "\n"));
#endif
}

void ensure_header(commands& cmdlist)
{
	if (cmdlist.empty() || cmdlist.begin()->compare(0, 3, _T("afx"))) {
		cmdlist.insert(cmdlist.begin(), _T("afx afxbkmk"));
	}
}

void add_item(commands& cmdlist, const tstring& line)
{
	commands::iterator i = cmdlist.begin();
	i ++;
	cmdlist.insert(i, line);
}

const TCHAR *basename(const TCHAR *name, int n)
{
	const TCHAR *p;
	
	if (n < 1 || ! name || ! *name) return name;
	p = name + _tcslen(name) - 1;
#ifdef NDEBUG
	for (; p > name; p = CharPrev(name, p)) {
		if (*p == _T('\\') && -- n == 0) return CharNext(p);
	}
#else
	for (; p > name; p --) {
		if (*p == _T('\\') && -- n == 0) return p + 1;
	}
#endif
	return p;
}

int islonum(int c) { return islower(c) || isdigit(c); }
int isupnum(int c) { return isupper(c) || isdigit(c); }

class sequence {
	typedef unsigned int bits;
	bits m_flag: 3;
	bits m_sequence;
public:
	sequence() : m_flag(0), m_sequence(0) {}
	bool is_enabled() const { return m_flag != 0; }
	void set_number0() { m_sequence = (m_sequence << 3) | 1; m_flag |= 1; }
	void set_number1() { m_sequence = (m_sequence << 3) | 5; m_flag |= 1; }
	void set_lower()   { m_sequence = (m_sequence << 3) | 2; m_flag |= 2; }
	void set_upper()   { m_sequence = (m_sequence << 3) | 6; m_flag |= 4; }
	void reverse()
	{
		bits seq = m_sequence, rev = 0;
		for (; seq != 0; seq >>= 3) {
			rev = (rev << 3) | (seq & 7);
		}
		m_sequence = rev;
	}
	TCHAR get_sequence_char() const
	{
		return _T(" 0a0A0a0")[m_flag];
	}
	void apply(commands&) const;
};

void sequence::apply(commands& cmdlist) const
{
	if (! is_enabled()) return;
	static int (*range[])(int) = {
		0, isdigit, islower, islonum, isupper, isupnum, isalpha, isalnum,
	};
	int (*isrange)(int) = range[m_flag];
	static TCHAR *begins = _T(" 0a  1A");
	static TCHAR *ends = _T(" 9z  9Z");
	TCHAR c, end;
	bits seq = m_sequence;
	c = begins[seq & 7];
	end = ends[seq & 7];
	seq >>= 3;
	for (commands::iterator i = cmdlist.begin(); i != cmdlist.end(); ++ i) {
		if ((*i)[0] != _T('"')) continue; // ignore no label
		if ((*i)[2] == _T(' ') && isrange((*i)[1])) {
			(*i)[1] = c;
			if (c == end) {
				if (seq == 0) seq = m_sequence;
				c = begins[seq & 7];
				end = ends[seq & 7];
				seq >>= 3;
			} else {
				c ++;
			}
		}
	}
}

struct options {
	unsigned int m_delete: 1;
	unsigned int m_move : 2;
	sequence m_sequence;
	const TCHAR *m_filename;
	tstring m_dirname, m_label;
	unsigned int m_execute;
	options(int argc, _TCHAR *argv[]);
};

options::options(int argc, _TCHAR *argv[])
	: m_delete(0), m_move(0), m_sequence(), m_filename(0),
	  m_dirname(), m_label(), m_execute(0)
{
	int params = 0, base = 0;
	bool label_with_path = false, exact_label = false;
	const TCHAR *dirname = _T("");
	
	for (int i = 1; i < argc; i ++) {
		if (argv[i][0] == _T('-')) {
			for (TCHAR *a = argv[i] + 1; *a; a ++) {
				if (*a == _T('d')) {
					m_delete = 1;
					m_move = 0;
				}
				else if (*a == _T('m')) {
					m_delete = 0;
					m_move = 1;
				}
				else if (*a == _T('M')) {
					m_delete = 0;
					m_move = 2;
				}
				else if (*a == _T('l')) {
					label_with_path = true;
				}
				else if (*a == _T('L')) {
					exact_label = true;
				}
				else if (*a == _T('n')) {
					m_sequence.set_number0();
				}
				else if (*a == _T('N')) {
					m_sequence.set_number1();
				}
				else if (*a == _T('a')) {
					m_sequence.set_lower();
				}
				else if (*a == _T('A')) {
					m_sequence.set_upper();
				}
				else if (*a == _T('e')) {
					m_execute = 1;
				}
				else if (*a >= _T('1') && *a <= _T('9')) {
					base = *a - _T('0');
				}
			}
		}
		else {
			switch (params ++) {
			case 0:
				m_filename = argv[i];
				break;
			case 1:
				dirname = argv[i];
				break;
			case 2:
				m_label = argv[i];
				break;
			default:
				if (! isspace(*argv[i])) m_label.append(_T(" "));
				m_label.append(argv[i]);
				break;
			}
		}
	}
	m_sequence.reverse();
	if (m_label.empty()) {
		m_label = basename(dirname, base);
		exact_label = false;
	}
	else if (label_with_path) {
		m_label.append(_T(" "));
		m_label.append(basename(dirname, base));
	}
	remove_double_quote(m_label);
	if (! exact_label && m_sequence.is_enabled()) {
		m_label.insert(0, _T("  "));
		m_label[0] = m_sequence.get_sequence_char();
	}
	add_double_quote(m_label);
	m_dirname = dirname;
	if (! m_dirname.empty()) {
		remove_double_quote(m_dirname);
		add_double_quote(m_dirname);
		if (m_execute == 1) {
			m_dirname.insert(0, _T("&EXEC "));
		} else {
			m_dirname.insert(0, _T("&CD "));
		}
	}
}

#ifndef NDEBUG
int main(int argc, TCHAR *argv[])
#else
#define argc __argc
#define argv __targv
int WINAPI _tWinMain(
	HINSTANCE hInstance, 
	HINSTANCE hPreInst, 
	LPTSTR lpszCmdLine, 
	int nCmdShow)
#endif
{
	setlocale(LC_ALL, "japanese");
	if (argc < 3) {
		static const TCHAR* err =
			_T("usage: afxbkmk [options] menu_file [folder [label]]");
#ifndef NDEBUG
		terr << err << endl;
#else
		MessageBox(NULL, err, _T("afxbkmk"), MB_OK);
#endif
		return 0;
	}
	options opts(argc, argv);
	commands cmdlist;
	if (! opts.m_sequence.is_enabled() && opts.m_dirname.empty()) {
		return 0;
	}
	bool find = load(opts.m_filename, opts.m_dirname, cmdlist, (opts.m_move != 2) ? NULL : &opts.m_label);
	ensure_header(cmdlist);
	if (opts.m_move == 2 && find) {
		opts.m_dirname.clear();
		opts.m_sequence.apply(cmdlist);
		add_item(cmdlist, opts.m_label);
		opts.m_label.clear();
	} else {
		if (! opts.m_delete && (! find || opts.m_move) && ! opts.m_dirname.empty()) {
			add_item(cmdlist, opts.m_label + _T("\t") + opts.m_dirname);
		}
		opts.m_dirname.clear();
		opts.m_label.clear();
		opts.m_sequence.apply(cmdlist);
	}
	save(opts.m_filename, cmdlist);
	return 0;
}
